/*****************************************************************************
    neterror.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <arpa/inet.h>

#include "params.h"
#include "globals.h"
#include "logfun.h"
#include "netio.h"
#include "util.h"
#include "neterror.h"

#if (RT_RRQ != 1) && (RT_WRQ != 2) && (RT_DATA != 3) && (RT_ACK != 4) && (RT_ERROR != 5) && (RT_OACK != 6)
#error RT_XXX definition mismatch
#endif


static char *txWhere(tftpR_t *request, int okMask)
{
    static char buf[16];
    static char *rtxt[] = { "GRQ","RRQ-","WRQ-","?" };
    static char *otxt[] = { "?","DATA","ACK", "INIT", 
                            "OACK","OACK/DATA", "OACK/ACK",""};

    strcpy(buf,   rtxt[request->reqType & 3]);
    strcpy(buf+4, otxt[okMask & 7]);
    return buf;
}


void logSequenceError(char *what, int seq, int awaiting)
{
    static char buf[64];

    if (what == NULL)
        what = _("packet");

    if (awaiting >= 0)
        snprintf(buf, sizeof(buf), " - awaiting %d", awaiting);
    else
        buf[0] = '\0';

    LOG_MSG2 (LOG_WARNING, _("Received out of sequence %s no. %d%s"), 
              what, seq, buf);
}

/*
	generic time-out check routine
*/

int retryPossible(tftpR_t *request, int errType, char *errText)
{
    if (errText == NULL)
        errText = "Too many retries";
    logRetryCount(request->retry);
    if (request->retry > request->times.retry) {
        logRetryError(NULL, request->retry);
        if (errType >= 0)
            sendError(request, errType, errText);
        return PC_ERROR;
    }
    request->retry++;
    return PC_RETRY;
}

/*
	generic receive packet check routine
*/

int checkRpack(tftpR_t *request, tftpD_t *packet, int packetLen, int okMask)
{
    char *where;
    int  noCount;
    int  maxLen;
    int  ptype;

    noCount = (okMask == BM_GRQ);	/* no retries on request packets */

/*
	system call errors

	there should not be any EAGAIN (signal) errors, because we do the
	recvfrom() only in follow-up of a preceeding select() call
*/
    if (packetLen < 0) {
        where = txWhere(request, okMask);
        logReceiveError(where);
        if (noCount) {
            sendError(request, PE_UNDEF, "Receive error");
            return PC_ERROR;
        }
        return retryPossible(request, PE_UNDEF, "Permanent receive error");
    }
    else if (packetLen < (int)sizeof(tftpH_t)) {
        where = txWhere(request, okMask);
        logShortPacket(packetLen, where);
        if (noCount)
            return PC_RETRY;
        return retryPossible(request, PE_UNDEF, "Short packet");
    }

    ptype = (int)ntohs(packet->head.opCode);

    switch (ptype) {
        case RT_ACK:
            maxLen = sizeof(tftpH_t);
            break;
        case RT_DATA:
            if ((request->option.types & (ROT_BSZ | ROT_BSZ2)))
                maxLen = sizeof(tftpH_t) + request->option.blocksize;
            else
                maxLen = sizeof(tftpD_t);
            break;
        default:
            maxLen = sizeof(tftpD_t);
            break;
    }
        
    if (packetLen > maxLen) {
        where = txWhere(request, okMask);
        logGiantPacket(packetLen, where);
        if (noCount)
            return PC_RETRY;
        return retryPossible(request, PE_UNDEF, "Giant packet");
    }

    switch (ptype) {
        case RT_RRQ:
        case RT_WRQ:
            okMask &= BM_GRQ;
            break;
        case RT_DATA:
            okMask &= BM_DATA;
            break;
        case RT_ACK:
            okMask &= BM_ACK;
            break;
        case RT_OACK:
            if (request->curBlock)
                okMask  = 0;
            else
                okMask &= BM_OACK;
            break;
        default:
            okMask  = 0;
            break;
    }

    if (okMask)
        return PC_OK;
    where = txWhere(request, okMask);
    if (ptype == RT_ERROR) {
        logPeerError(ntohs(packet->head.errCode),getErrText(packet,packetLen));
        return PC_ERROR;
    }
    else {
        logTypeError(ptype, where);
    }
    if (noCount)
        return PC_UNKP;

    return retryPossible(request, PE_UNDEF, "Bad packet");
}


int checkSelError(tftpR_t *request, int rtc, int okMask)
{
    char *where;
    struct timeval timer = { 0 , 0 };

    if (rtc > 0)
        return PC_OK;

    where = txWhere(request, okMask);

    if (rtc < 0) {
        if ((errno == EAGAIN) || (errno == EINTR))
            return PC_AGAIN;
        logSelError(where);
        return PC_ERROR;
    }

    /* here: rtc == 0 i.e.: timeout */

    if (okMask == BM_GRQ) 
        return PC_RETRY;

    if (okMask & BM_ACK) {
        if (request->reqType == RT_RRQ)
            timer = (request->reqRole == RR_SERVER) ? request->times.sack :
                                                      request->times.rack;
        else
            timer = (request->reqRole == RR_SERVER) ? request->times.rack :
                                                      request->times.sack;
    }
    else if (okMask & BM_DATA) {
        if (request->reqType == RT_RRQ)
            timer = (request->reqRole == RR_SERVER) ? request->times.send :
                                                      request->times.recv;
        else
            timer = (request->reqRole == RR_SERVER) ? request->times.recv :
                                                      request->times.send;
    }
    logSelTimeout(where, request->retry, &timer);
    return retryPossible(request, PE_UNDEF, "Too many retries");
}

int checkSendError(tftpR_t *request, int rtc, int len, int okMask)
{
    char *where;

    if (rtc == len)
       return PC_OK; 

    where = txWhere(request, okMask);
    logSendError(where);
    return retryPossible(request, -1, NULL);
}

